from abc import ABCMeta, abstractmethod
from typing import Any
import uuid


class Message(object):
    def __init__(self, data: Any = None):
        self.uid = uuid.uuid4().int
        self.data = data

    def getUID(self) -> int:
        return self.uid


class Observable(object):
    """Base class for the objects that need to notify
    other objects

    """

    def __init__(self):
        self.__observers = []

    def addObserver(self, obs: "Observer"):
        """Adds an observer. The *obs* must derive from the :class:`Observer` base class

        Args:
          obs
            Instance of a subclass of :class:`Observer` base class

        """
        self.__observers.append(obs)

    def notifyObservers(self, msg: Message):
        """Call this method to send data to all the observers

        Args:
          msg
            Message to be sent

        """
        for obs in self.__observers:
            if not obs.isMessageHandeld(msg):
                obs.setLastMessageReceivedUID(msg)
                obs.handleObservableMsg(sender=self, msg=msg)


class Observer(metaclass=ABCMeta):
    """Abstract base class for the objects that need to be notified"""

    def __init__(self):
        self.__last_msg_received_id = -1

    def setLastMessageReceivedUID(self, msg: Message):
        self.__last_msg_received_id = msg.getUID()

    def getLastMessageReceivedUID(self) -> int:
        return self.__last_msg_received_id

    def isMessageHandeld(self, msg: Message) -> bool:
        return msg.getUID() == self.getLastMessageReceivedUID()

    @abstractmethod
    def handleObservableMsg(self, sender: Observable, msg: Message):
        """Method called be the observable each time they notify the observers

        Args:
          sender
            :class:`Observable` object that sents the notification
          msg
            Message received

        """
        pass


if __name__ == "__main__":
    import doctest

    doctest.testmod()
